function createPromise() {
	let promise, resolve, reject;

	promise = new Promise((a, b) => {
		resolve = a;
		reject = b;
	});

	return {promise, resolve, reject}
}

function isNode() {
	if (typeof window !== "undefined") {
		return false
	}

	if (typeof process === "undefined") {
		return false
	}

	return true
}

function node_cleanupHandler(dependencies, global_vars) {
	const {fs} = dependencies;

	while (global_vars.cleanup_items.length) {
		const item = global_vars.cleanup_items.shift();

		try {
			fs.unlinkSync(item);
		} catch {}
	}
}

async function createTemporaryResourceImplementation(
	{environment, dependencies, global_vars}, data, options
) {
	if (environment === "node") {
		const {fs, os, path, crypto, process} = dependencies;

		if (!global_vars.cleanupHandler_set) {
			process.on("SIGTERM", () => {
				node_cleanupHandler(dependencies, global_vars);
			});

			process.on("exit", () => {
				node_cleanupHandler(dependencies, global_vars);
			});

			global_vars.cleanupHandler_set = true;
		}

		const tmpname = crypto.randomBytes(16).toString("hex");
		let extension = ".txt";

		if (options.type === "text/javascript") {
			extension = ".mjs";
		}

		const tmppath = path.join(os.tmpdir(), tmpname + extension);

		fs.writeFileSync(tmppath, data);

		const location = fs.realpathSync(tmppath);

		if (options.auto_cleanup) {
			global_vars.cleanup_items.push(location);
		}

		return {
			location,
			cleanup() {
				try {
					fs.unlinkSync(location);
				} catch {}
			}
		}
	} else {
		let blob = new Blob([data], {type: options.type});
		const location = URL.createObjectURL(blob);

		return {
			location,
			cleanup() {
				URL.revokeObjectURL(location);
			}
		}
	}
}

async function importDependenciesForEnvironment(environment) {
	if (environment !== "node") return {}

	let node_modules = {};

	const dependencies = ["fs", "os", "path", "crypto", "process"];

	for (const dependency of dependencies) {
		node_modules[dependency] = (await import(`node:${dependency}`)).default;
	}

	return node_modules
}

function getInitialStateForEnvironment(environment) {
	if (environment !== "node") return {}

	return {
		cleanupHandler_set: false,
		cleanup_items: []
	}
}

function createTemporaryResourceFactory({node = true} = {}) {
	const environment = node ? "node" : "browser";
	const global_data_symbol = Symbol.for(`@anio-js-foundation/create-temporary-resource-factory-${environment}`);

	if (!(global_data_symbol in globalThis)) {
		globalThis[global_data_symbol] = getInitialStateForEnvironment(environment);
	}

	const global_vars = globalThis[global_data_symbol];

	return async (data, {type = "text/plain", auto_cleanup = true} = {}) => {
		const dependencies = await importDependenciesForEnvironment(environment);

		return await createTemporaryResourceImplementation({
			environment, dependencies, global_vars
		}, data, {type, auto_cleanup})
	}
}

const createTemporaryResource = createTemporaryResourceFactory({
	node: isNode()
});

function eventEmitter(valid_events) {
	let methods = {};
	let registered_handler = {};
	let event_handler_added_handler = null;
	let event_dispatched_handler = null;
	let event_handler_removed_handler = null;

	for (let event of valid_events) {
		registered_handler[event] = [];
	}

	const handlerInstalledForEvent = (event_name, fn) => {
		return registered_handler[event_name].filter(handler => {
			return Object.is(handler, fn)
		}).length > 0
	};

	methods.on = function on(event_name, fn) {
		return methods.addEventListener(event_name, fn)
	};

	methods.addEventListener = function addEventListener(event_name, fn) {
		if (!(event_name in registered_handler)) {
			throw new Error(`Invalid event '${event_name}'.`)
		} else if (handlerInstalledForEvent(event_name, fn)) {
			throw new Error(`This *exact* function was already registered for this event.`)
		}

		registered_handler[event_name].push(fn);

		if (typeof event_handler_added_handler === "function") {
			event_handler_added_handler(event_name, fn);
		}
	};

	methods.removeEventListener = function removeEventListener(event_name, fn) {
		if (!(event_name in registered_handler)) {
			throw new Error(`Invalid event '${event_name}'.`)
		} else if (!handlerInstalledForEvent(event_name, fn)) {
			throw new Error(`This function was not registered for this event.`)
		}

		registered_handler[event_name] = registered_handler[event_name].filter(handler => {
			return !Object.is(handler, fn)
		});

		if (typeof event_handler_removed_handler === "function") {
			event_handler_removed_handler(event_name, fn);
		}
	};

	return {
		setOnEventHandlerAddedHandler(fn) {
			event_handler_added_handler = fn;
		},

		setOnEventDispatchedHandler(fn) {
			event_dispatched_handler = fn;
		},

		setOnEventHandlerRemovedHandler(fn) {
			event_handler_removed_handler = fn;
		},

		install(obj) {
			for (const method of ["on", "addEventListener", "removeEventListener"]) {
				obj[method] = methods[method];
			}

			return function dispatchEvent(event_name, ...event_data) {
				if (!(event_name in registered_handler)) {
					throw new Error(`Invalid event '${event_name}'.`)
				}

				//
				// todo: maybe add "blocking" events
				// i.e. each event handler gets executed consecutively
				// example of code flow:
				// await handler()
				// await next_handler()
				// await all_handlers_dispatched()
				//
				for (const handler of registered_handler[event_name]) {
					setTimeout(handler, 0, ...event_data);
				}

				if (typeof event_dispatched_handler === "function") {
					setTimeout(event_dispatched_handler, 0, event_name, event_data);
				}
			}
		}
	}
}

const global_uuid = `FACD900A-9868-4CCB-B660-9DDF52D1BE12`;

const symbol = Symbol.for(
	`@anio-js-foundation/anio-global-store-${global_uuid}`
);

//
// return the global store
// create it if it doesn't exist yet
//
function getGlobalStore() {
	if (symbol in globalThis) {
		return globalThis[symbol]
	}

	globalThis[symbol] = {
		//
		// private_items will never be
		// passed down to a worker created 
		// by @anio-js-foundation/node-create-worker
		//
		private_items: new Map(),
		//
		// shared_items will be passed
		// down to a worker created
		// by @anio-js-foundation/node-create-worker
		//
		shared_items: new Map()
	};

	return globalThis[symbol]
}

function genericSetItem(map, key, value) {
	const previous_item_value = map.has(key) ? map.get(key) : null;

	map.set(key, value);

	return previous_item_value
}

function genericGetItem(map, key) {
	if (!map.has(key)) {
		throw new Error(`No such key '${key}' in anio-global-store. (operation: get)`)
	}

	return map.get(key)
}

function genericDeleteItem(map, key) {
	if (!map.has(key)) {
		throw new Error(`No such key '${key}' in anio-global-store. (operation: delete)`)
	}

	map.delete(key);
}



//
// private globals
//
function hasItem(key) {
	const {private_items} = getGlobalStore();

	return private_items.has(key)
}

function setItem(key, value) {
	const {private_items} = getGlobalStore();

	return genericSetItem(private_items, key, value)
}

function getItem(key) {
	const {private_items} = getGlobalStore();

	return genericGetItem(private_items, key)
}

function deleteItem(key) {
	const {private_items} = getGlobalStore();

	genericDeleteItem(private_items, key);
}

function getItems() {
	const {private_items} = getGlobalStore();

	// only return a copy
	return new Map(private_items)
}

//
// "shared" globals
//
function hasSharedItem(key) {
	const {shared_items} = getGlobalStore();

	return shared_items.has(key)
}

function setSharedItem(key, value) {
	const {shared_items} = getGlobalStore();

	return genericSetItem(shared_items, key, value)
}

function getSharedItem(key) {
	const {shared_items} = getGlobalStore();

	return genericGetItem(shared_items, key)
}

function deleteSharedItem(key) {
	const {shared_items} = getGlobalStore();

	genericDeleteItem(shared_items, key);
}

function getSharedItems() {
	const {shared_items} = getGlobalStore();

	return new Map(shared_items)
}

//
var index = {
	hasItem,
	setItem,
	getItem,
	deleteItem,
	getItems,

	hasSharedItem,
	setSharedItem,
	getSharedItem,
	deleteSharedItem,
	getSharedItems
};

var bootstrap_code = "import process$1 from 'node:process';\n\nfunction nodeIsProcessRunning(pid) {\n\ttry {\n\t\treturn process$1.kill(pid, 0)\n\t} catch (e) {\n\t\treturn e.code !== \"ESRCH\"\n\t}\n}\n\nfunction eventEmitter(valid_events) {\n\tlet methods = {};\n\tlet registered_handler = {};\n\tlet event_handler_added_handler = null;\n\tlet event_dispatched_handler = null;\n\tlet event_handler_removed_handler = null;\n\n\tfor (let event of valid_events) {\n\t\tregistered_handler[event] = [];\n\t}\n\n\tconst handlerInstalledForEvent = (event_name, fn) => {\n\t\treturn registered_handler[event_name].filter(handler => {\n\t\t\treturn Object.is(handler, fn)\n\t\t}).length > 0\n\t};\n\n\tmethods.on = function on(event_name, fn) {\n\t\treturn methods.addEventListener(event_name, fn)\n\t};\n\n\tmethods.addEventListener = function addEventListener(event_name, fn) {\n\t\tif (!(event_name in registered_handler)) {\n\t\t\tthrow new Error(`Invalid event '${event_name}'.`)\n\t\t} else if (handlerInstalledForEvent(event_name, fn)) {\n\t\t\tthrow new Error(`This *exact* function was already registered for this event.`)\n\t\t}\n\n\t\tregistered_handler[event_name].push(fn);\n\n\t\tif (typeof event_handler_added_handler === \"function\") {\n\t\t\tevent_handler_added_handler(event_name, fn);\n\t\t}\n\t};\n\n\tmethods.removeEventListener = function removeEventListener(event_name, fn) {\n\t\tif (!(event_name in registered_handler)) {\n\t\t\tthrow new Error(`Invalid event '${event_name}'.`)\n\t\t} else if (!handlerInstalledForEvent(event_name, fn)) {\n\t\t\tthrow new Error(`This function was not registered for this event.`)\n\t\t}\n\n\t\tregistered_handler[event_name] = registered_handler[event_name].filter(handler => {\n\t\t\treturn !Object.is(handler, fn)\n\t\t});\n\n\t\tif (typeof event_handler_removed_handler === \"function\") {\n\t\t\tevent_handler_removed_handler(event_name, fn);\n\t\t}\n\t};\n\n\treturn {\n\t\tsetOnEventHandlerAddedHandler(fn) {\n\t\t\tevent_handler_added_handler = fn;\n\t\t},\n\n\t\tsetOnEventDispatchedHandler(fn) {\n\t\t\tevent_dispatched_handler = fn;\n\t\t},\n\n\t\tsetOnEventHandlerRemovedHandler(fn) {\n\t\t\tevent_handler_removed_handler = fn;\n\t\t},\n\n\t\tinstall(obj) {\n\t\t\tfor (const method of [\"on\", \"addEventListener\", \"removeEventListener\"]) {\n\t\t\t\tobj[method] = methods[method];\n\t\t\t}\n\n\t\t\treturn function dispatchEvent(event_name, ...event_data) {\n\t\t\t\tif (!(event_name in registered_handler)) {\n\t\t\t\t\tthrow new Error(`Invalid event '${event_name}'.`)\n\t\t\t\t}\n\n\t\t\t\t//\n\t\t\t\t// todo: maybe add \"blocking\" events\n\t\t\t\t// i.e. each event handler gets executed consecutively\n\t\t\t\t// example of code flow:\n\t\t\t\t// await handler()\n\t\t\t\t// await next_handler()\n\t\t\t\t// await all_handlers_dispatched()\n\t\t\t\t//\n\t\t\t\tfor (const handler of registered_handler[event_name]) {\n\t\t\t\t\tsetTimeout(handler, 0, ...event_data);\n\t\t\t\t}\n\n\t\t\t\tif (typeof event_dispatched_handler === \"function\") {\n\t\t\t\t\tsetTimeout(event_dispatched_handler, 0, event_name, event_data);\n\t\t\t\t}\n\t\t\t}\n\t\t}\n\t}\n}\n\nconst global_uuid = `FACD900A-9868-4CCB-B660-9DDF52D1BE12`;\n\nconst symbol = Symbol.for(\n\t`@anio-js-foundation/anio-global-store-${global_uuid}`\n);\n\n//\n// return the global store\n// create it if it doesn't exist yet\n//\nfunction getGlobalStore() {\n\tif (symbol in globalThis) {\n\t\treturn globalThis[symbol]\n\t}\n\n\tglobalThis[symbol] = {\n\t\t//\n\t\t// private_items will never be\n\t\t// passed down to a worker created \n\t\t// by @anio-js-foundation/node-create-worker\n\t\t//\n\t\tprivate_items: new Map(),\n\t\t//\n\t\t// shared_items will be passed\n\t\t// down to a worker created\n\t\t// by @anio-js-foundation/node-create-worker\n\t\t//\n\t\tshared_items: new Map()\n\t};\n\n\treturn globalThis[symbol]\n}\n\nfunction genericSetItem(map, key, value) {\n\tconst previous_item_value = map.has(key) ? map.get(key) : null;\n\n\tmap.set(key, value);\n\n\treturn previous_item_value\n}\n\nfunction setSharedItem(key, value) {\n\tconst {shared_items} = getGlobalStore();\n\n\treturn genericSetItem(shared_items, key, value)\n}\n\nlet global_parent_pid = null;\n\nfunction exitIfParentIsDead() {\n\tif (global_parent_pid === null) return\n\n\tif (!nodeIsProcessRunning(global_parent_pid)) {\n\t\tprocess.exit(1);\n\n\t\treturn true\n\t}\n\n\treturn false\n}\n\nfunction exitIfParentIsDeadLoop() {\n\tif (exitIfParentIsDead()) return\n\n\tsetTimeout(exitIfParentIsDeadLoop, 250);\n}\n\nfunction createWorkerThis() {\n\tlet new_this = {};\n\n\tlet event_emitter = eventEmitter([\"message\"]);\n\n\tlet dispatchEvent = event_emitter.install(new_this);\n\n\tprocess.on(\"message\", msg => {\n\t\tdispatchEvent(\"message\", msg);\n\t});\n\n\tObject.defineProperty(new_this, \"sendMessage\", {\n\t\tenumerable: true,\n\n\t\tset() { throw new Error(`Cannot set sendMessage.`) },\n\t\tget() {\n\t\t\treturn (str) => {\n\t\t\t\texitIfParentIsDead();\n\n\t\t\t\treturn process.send(str)\n\t\t\t}\n\t\t}\n\t});\n\n\treturn new_this\n}\n\nlet onMessageHandler = (msg) => {\n\tif (msg.startsWith(\"init\")) {\n\t\tprocess.removeListener(\"message\", onMessageHandler);\n\n\t\tconst payload = JSON.parse(msg.slice(\"init\".length));\n\n\t\tglobal_parent_pid = payload.parent_pid;\n\n\t\timport(payload.worker_file_path)\n\t\t.then(mod => {\n\t\t\tconst init_args = payload.worker_args;\n\n\t\t\t//\n\t\t\t// restore shared items from anio-global-store\n\t\t\t//\n\t\t\tconst {anio_global_store_shared_items} = payload;\n\n\t\t\tfor (const key in anio_global_store_shared_items) {\n\t\t\t\tsetSharedItem(key, anio_global_store_shared_items[key]);\n\t\t\t}\n\n\t\t\tlet new_this = createWorkerThis();\n\n\t\t\treturn mod.NodeWorkerMain.apply(new_this, init_args)\n\t\t})\n\t\t.then(() => {\n\t\t\tprocess.send(payload.init_token);\n\n\t\t\tsetTimeout(exitIfParentIsDeadLoop, 0);\n\t\t});\n\t}\n};\n\nprocess.on(\"message\", onMessageHandler);\n";

async function createNodeWorkerProcess(dependencies, options) {
	const {spawn} = dependencies;

	let node_binary_path = "node", silent = true;

	if ("node_binary" in options) {
		node_binary_path = options.node_binary;
	}

	if ("silent" in options) {
		silent = options.silent;
	}

	const stdio = silent ? ["pipe", "pipe", "pipe", "ipc"] : ["pipe", "inherit", "inherit", "ipc"];

	const bootstrap = await createTemporaryResource(
		bootstrap_code, {type: "text/javascript"}
	);

	const child = spawn(node_binary_path, [
		bootstrap.location
	], {
		stdio
	});

	// delete bootstrap file after first ipc message
	let onMessageHandler = (msg) => {
		child.removeListener("message", onMessageHandler);

		bootstrap.cleanup();
	};

	child.on("message", onMessageHandler);

	return child
}

function createWorkerInstance({
	child,
	worker_message_buffer
}) {
	let instance = {};
	let event_emitter = eventEmitter(["message"]);

	let dispatchEvent = event_emitter.install(instance);

	child.on("message", msg => {
		dispatchEvent("message", msg);
	});

	/* dump message buffer on first 'message' handler installed */
	event_emitter.setOnEventHandlerAddedHandler((event_name, handler) => {
		if (event_name !== "message") return

		// dump buffer
		while (worker_message_buffer.length) {
			const msg = worker_message_buffer.shift();

			handler(msg);
		}
	});

	Object.defineProperty(instance, "sendMessage", {
		enumerable: true,
		get() { return (str) => child.send(str) }
	});

	Object.defineProperty(instance, "terminate", {
		enumerable: true,
		get() { return () => child.kill("SIGKILL") }
	});

	return instance
}

async function nodeCreateWorkerImplementation(dependencies, worker_file_path, worker_args, additional = {}) {
	let {promise, resolve, reject} = createPromise();

	const init_token = Math.random().toString(32) + "_" + Math.random().toString(32);

	let child = await createNodeWorkerProcess(dependencies, additional);
	let worker_message_buffer = [];

	child.on("error", reject);

	child.on("message", msg => {
		if (msg === init_token) {
			resolve(
				createWorkerInstance({
					child,
					worker_message_buffer
				})
			);
		}
		// buffer other messages between worker and main script
		else {
			worker_message_buffer.push(msg);
		}
	});

	const anio_global_store_shared_items = index.getSharedItems();

	child.send("init" + JSON.stringify({
		parent_pid: process.pid,
		worker_file_path,
		worker_args,
		init_token,
		additional,
		//
		// pass down shared items from "anio-global-store"
		//
		anio_global_store_shared_items: Object.fromEntries(anio_global_store_shared_items)
	}));

	return promise
}

async function nodeCreateWorker(...args) {
	const {default: child_process} = await import('node:child_process');
	const {spawn} = child_process;

	const dependencies = {spawn};

	return await nodeCreateWorkerImplementation(
		dependencies, ...args
	)
}

export { nodeCreateWorker as default };
